import unittest
import os
import sys
from fontTools import t1Lib
from fontTools.pens.basePen import NullPen
from fontTools.misc.psCharStrings import T1CharString
import random


CWD = os.path.abspath(os.path.dirname(__file__))
DATADIR = os.path.join(CWD, 'data')
# I used `tx` to convert PFA to LWFN (stored in the data fork)
LWFN = os.path.join(DATADIR, 'TestT1-Regular.lwfn')
PFA = os.path.join(DATADIR, 'TestT1-Regular.pfa')
PFB = os.path.join(DATADIR, 'TestT1-Regular.pfb')
WEIRD_ZEROS = os.path.join(DATADIR, 'TestT1-weird-zeros.pfa')
# ellipsis is hinted with 55 131 296 131 537 131 vstem3 0 122 hstem
ELLIPSIS_HINTED = os.path.join(DATADIR, 'TestT1-ellipsis-hinted.pfa')


class FindEncryptedChunksTest(unittest.TestCase):

	def test_findEncryptedChunks(self):
		with open(PFA, "rb") as f:
			data = f.read()
		chunks = t1Lib.findEncryptedChunks(data)
		self.assertEqual(len(chunks), 3)
		self.assertFalse(chunks[0][0])
		# the second chunk is encrypted
		self.assertTrue(chunks[1][0])
		self.assertFalse(chunks[2][0])

	def test_findEncryptedChunks_weird_zeros(self):
		with open(WEIRD_ZEROS, 'rb') as f:
			data = f.read()

		# Just assert that this doesn't raise any exception for not finding the
		# end of eexec
		t1Lib.findEncryptedChunks(data)


class DecryptType1Test(unittest.TestCase):

	def test_decryptType1(self):
		with open(PFA, "rb") as f:
			data = f.read()
		decrypted = t1Lib.decryptType1(data)
		self.assertNotEqual(decrypted, data)


class ReadWriteTest(unittest.TestCase):

	def test_read_pfa_write_pfb(self):
		font = t1Lib.T1Font(PFA)
		data = self.write(font, 'PFB')
		self.assertEqual(font.getData(), data)

	def test_read_and_parse_pfa_write_pfb(self):
		font = t1Lib.T1Font(PFA)
		font.parse()
		saved_font = self.write(font, 'PFB', dohex=False, doparse=True)
		self.assertTrue(same_dicts(font.font, saved_font))

	def test_read_pfb_write_pfa(self):
		font = t1Lib.T1Font(PFB)
		# 'OTHER' == 'PFA'
		data = self.write(font, 'OTHER', dohex=True)
		self.assertEqual(font.getData(), data)

	def test_read_and_parse_pfb_write_pfa(self):
		font = t1Lib.T1Font(PFB)
		font.parse()
		# 'OTHER' == 'PFA'
		saved_font = self.write(font, 'OTHER', dohex=True, doparse=True)
		self.assertTrue(same_dicts(font.font, saved_font))

	def test_read_with_path(self):
		import pathlib
		font = t1Lib.T1Font(pathlib.Path(PFB))

	@staticmethod
	def write(font, outtype, dohex=False, doparse=False):
		temp = os.path.join(DATADIR, 'temp.' + outtype.lower())
		try:
			font.saveAs(temp, outtype, dohex=dohex)
			newfont = t1Lib.T1Font(temp)
			if doparse:
				newfont.parse()
				data = newfont.font
			else:
				data = newfont.getData()
		finally:
			if os.path.exists(temp):
				os.remove(temp)
		return data


class T1FontTest(unittest.TestCase):

	def test_parse_lwfn(self):
		# the extended attrs are lost on git so we can't auto-detect 'LWFN'
		font = t1Lib.T1Font(LWFN, kind="LWFN")
		font.parse()
		self.assertEqual(font['FontName'], 'TestT1-Regular')
		self.assertTrue('Subrs' in font['Private'])

	def test_parse_pfa(self):
		font = t1Lib.T1Font(PFA)
		font.parse()
		self.assertEqual(font['FontName'], 'TestT1-Regular')
		self.assertTrue('Subrs' in font['Private'])

	def test_parse_pfb(self):
		font = t1Lib.T1Font(PFB)
		font.parse()
		self.assertEqual(font['FontName'], 'TestT1-Regular')
		self.assertTrue('Subrs' in font['Private'])

	def test_getGlyphSet(self):
		font = t1Lib.T1Font(PFA)
		glyphs = font.getGlyphSet()
		i = random.randrange(len(glyphs))
		aglyph = list(glyphs.values())[i]
		self.assertTrue(hasattr(aglyph, 'draw'))
		self.assertFalse(hasattr(aglyph, 'width'))
		aglyph.draw(NullPen())
		self.assertTrue(hasattr(aglyph, 'width'))


class EditTest(unittest.TestCase):

	def test_edit_pfa(self):
		font = t1Lib.T1Font(PFA)
		ellipsis = font.getGlyphSet()["ellipsis"]
		ellipsis.decompile()
		program = []
		for v in ellipsis.program:
			try:
				program.append(int(v))
			except:
				program.append(v)
				if v == 'hsbw':
					hints = [55, 131, 296, 131, 537, 131, 'vstem3', 0, 122, 'hstem']
					program.extend(hints)
		ellipsis.program = program
		# 'OTHER' == 'PFA'
		saved_font = self.write(font, 'OTHER', dohex=True, doparse=True)
		hinted_font = t1Lib.T1Font(ELLIPSIS_HINTED)
		hinted_font.parse()
		self.assertTrue(same_dicts(hinted_font.font, saved_font))

	@staticmethod
	def write(font, outtype, dohex=False, doparse=False):
		temp = os.path.join(DATADIR, 'temp.' + outtype.lower())
		try:
			font.saveAs(temp, outtype, dohex=dohex)
			newfont = t1Lib.T1Font(temp)
			if doparse:
				newfont.parse()
				data = newfont.font
			else:
				data = newfont.getData()
		finally:
			if os.path.exists(temp):
				os.remove(temp)
		return data


def same_dicts(dict1, dict2):
	if dict1.keys() != dict2.keys():
		return False
	for key, value in dict1.items():
		if isinstance(value, dict):
			if not same_dicts(value, dict2[key]):
				return False
		elif isinstance(value, list):
			if len(value) != len(dict2[key]):
				return False
			for elem1, elem2 in zip(value, dict2[key]):
				if isinstance(elem1, T1CharString):
					elem1.compile()
					elem2.compile()
					if elem1.bytecode != elem2.bytecode:
						return False
				else:
					if elem1 != elem2:
						return False
		elif isinstance(value, T1CharString):
			value.compile()
			dict2[key].compile()
			if value.bytecode != dict2[key].bytecode:
				return False
		else:
			if value != dict2[key]:
				return False
	return True


if __name__ == '__main__':
	import sys
	sys.exit(unittest.main())



# Copyright 2013 Google, Inc. All Rights Reserved.
#
# Google Author(s): Behdad Esfahbod, Roozbeh Pournader

from fontTools import ttLib
from fontTools.ttLib.tables.DefaultTable import DefaultTable
from fontTools.ttLib.tables import otTables
from fontTools.merge.base import add_method, mergeObjects
from fontTools.merge.util import *
import logging


log = logging.getLogger("fontTools.merge")


def mergeLookupLists(lst):
	# TODO Do smarter merge.
	return sumLists(lst)

def mergeFeatures(lst):
	assert lst
	self = otTables.Feature()
	self.FeatureParams = None
	self.LookupListIndex = mergeLookupLists([l.LookupListIndex for l in lst if l.LookupListIndex])
	self.LookupCount = len(self.LookupListIndex)
	return self

def mergeFeatureLists(lst):
	d = {}
	for l in lst:
		for f in l:
			tag = f.FeatureTag
			if tag not in d:
				d[tag] = []
			d[tag].append(f.Feature)
	ret = []
	for tag in sorted(d.keys()):
		rec = otTables.FeatureRecord()
		rec.FeatureTag = tag
		rec.Feature = mergeFeatures(d[tag])
		ret.append(rec)
	return ret

def mergeLangSyses(lst):
	assert lst

	# TODO Support merging ReqFeatureIndex
	assert all(l.ReqFeatureIndex == 0xFFFF for l in lst)

	self = otTables.LangSys()
	self.LookupOrder = None
	self.ReqFeatureIndex = 0xFFFF
	self.FeatureIndex = mergeFeatureLists([l.FeatureIndex for l in lst if l.FeatureIndex])
	self.FeatureCount = len(self.FeatureIndex)
	return self

def mergeScripts(lst):
	assert lst

	if len(lst) == 1:
		return lst[0]
	langSyses = {}
	for sr in lst:
		for lsr in sr.LangSysRecord:
			if lsr.LangSysTag not in langSyses:
				langSyses[lsr.LangSysTag] = []
			langSyses[lsr.LangSysTag].append(lsr.LangSys)
	lsrecords = []
	for tag, langSys_list in sorted(langSyses.items()):
		lsr = otTables.LangSysRecord()
		lsr.LangSys = mergeLangSyses(langSys_list)
		lsr.LangSysTag = tag
		lsrecords.append(lsr)

	self = otTables.Script()
	self.LangSysRecord = lsrecords
	self.LangSysCount = len(lsrecords)
	dfltLangSyses = [s.DefaultLangSys for s in lst if s.DefaultLangSys]
	if dfltLangSyses:
		self.DefaultLangSys = mergeLangSyses(dfltLangSyses)
	else:
		self.DefaultLangSys = None
	return self

def mergeScriptRecords(lst):
	d = {}
	for l in lst:
		for s in l:
			tag = s.ScriptTag
			if tag not in d:
				d[tag] = []
			d[tag].append(s.Script)
	ret = []
	for tag in sorted(d.keys()):
		rec = otTables.ScriptRecord()
		rec.ScriptTag = tag
		rec.Script = mergeScripts(d[tag])
		ret.append(rec)
	return ret

otTables.ScriptList.mergeMap = {
	'ScriptCount': lambda lst: None, # TODO
	'ScriptRecord': mergeScriptRecords,
}
otTables.BaseScriptList.mergeMap = {
	'BaseScriptCount': lambda lst: None, # TODO
	# TODO: Merge duplicate entries
	'BaseScriptRecord': lambda lst: sorted(sumLists(lst), key=lambda s: s.BaseScriptTag),
}

otTables.FeatureList.mergeMap = {
	'FeatureCount': sum,
	'FeatureRecord': lambda lst: sorted(sumLists(lst), key=lambda s: s.FeatureTag),
}

otTables.LookupList.mergeMap = {
	'LookupCount': sum,
	'Lookup': sumLists,
}

otTables.Coverage.mergeMap = {
	'Format': min,
	'glyphs': sumLists,
}

otTables.ClassDef.mergeMap = {
	'Format': min,
	'classDefs': sumDicts,
}

otTables.LigCaretList.mergeMap = {
	'Coverage': mergeObjects,
	'LigGlyphCount': sum,
	'LigGlyph': sumLists,
}

otTables.AttachList.mergeMap = {
	'Coverage': mergeObjects,
	'GlyphCount': sum,
	'AttachPoint': sumLists,
}

# XXX Renumber MarkFilterSets of lookups
otTables.MarkGlyphSetsDef.mergeMap = {
	'MarkSetTableFormat': equal,
	'MarkSetCount': sum,
	'Coverage': sumLists,
}

otTables.Axis.mergeMap = {
	'*': mergeObjects,
}

# XXX Fix BASE table merging
otTables.BaseTagList.mergeMap = {
	'BaseTagCount': sum,
	'BaselineTag': sumLists,
}

otTables.GDEF.mergeMap = \
otTables.GSUB.mergeMap = \
otTables.GPOS.mergeMap = \
otTables.BASE.mergeMap = \
otTables.JSTF.mergeMap = \
otTables.MATH.mergeMap = \
{
	'*': mergeObjects,
	'Version': max,
}

ttLib.getTableClass('GDEF').mergeMap = \
ttLib.getTableClass('GSUB').mergeMap = \
ttLib.getTableClass('GPOS').mergeMap = \
ttLib.getTableClass('BASE').mergeMap = \
ttLib.getTableClass('JSTF').mergeMap = \
ttLib.getTableClass('MATH').mergeMap = \
{
	'tableTag': onlyExisting(equal), # XXX clean me up
	'table': mergeObjects,
}

@add_method(ttLib.getTableClass('GSUB'))
def merge(self, m, tables):

	assert len(tables) == len(m.duplicateGlyphsPerFont)
	for i,(table,dups) in enumerate(zip(tables, m.duplicateGlyphsPerFont)):
		if not dups: continue
		if table is None or table is NotImplemented:
			log.warning("Have non-identical duplicates to resolve for '%s' but no GSUB. Are duplicates intended?: %s", m.fonts[i]._merger__name, dups)
			continue

		synthFeature = None
		synthLookup = None
		for script in table.table.ScriptList.ScriptRecord:
			if script.ScriptTag == 'DFLT': continue # XXX
			for langsys in [script.Script.DefaultLangSys] + [l.LangSys for l in script.Script.LangSysRecord]:
				if langsys is None: continue # XXX Create!
				feature = [v for v in langsys.FeatureIndex if v.FeatureTag == 'locl']
				assert len(feature) <= 1
				if feature:
					feature = feature[0]
				else:
					if not synthFeature:
						synthFeature = otTables.FeatureRecord()
						synthFeature.FeatureTag = 'locl'
						f = synthFeature.Feature = otTables.Feature()
						f.FeatureParams = None
						f.LookupCount = 0
						f.LookupListIndex = []
						table.table.FeatureList.FeatureRecord.append(synthFeature)
						table.table.FeatureList.FeatureCount += 1
					feature = synthFeature
					langsys.FeatureIndex.append(feature)
					langsys.FeatureIndex.sort(key=lambda v: v.FeatureTag)

				if not synthLookup:
					subtable = otTables.SingleSubst()
					subtable.mapping = dups
					synthLookup = otTables.Lookup()
					synthLookup.LookupFlag = 0
					synthLookup.LookupType = 1
					synthLookup.SubTableCount = 1
					synthLookup.SubTable = [subtable]
					if table.table.LookupList is None:
						# mtiLib uses None as default value for LookupList,
						# while feaLib points to an empty array with count 0
						# TODO: make them do the same
						table.table.LookupList = otTables.LookupList()
						table.table.LookupList.Lookup = []
						table.table.LookupList.LookupCount = 0
					table.table.LookupList.Lookup.append(synthLookup)
					table.table.LookupList.LookupCount += 1

				if feature.Feature.LookupListIndex[:1] != [synthLookup]:
					feature.Feature.LookupListIndex[:0] = [synthLookup]
					feature.Feature.LookupCount += 1

	DefaultTable.merge(self, m, tables)
	return self

@add_method(otTables.SingleSubst,
		otTables.MultipleSubst,
		otTables.AlternateSubst,
		otTables.LigatureSubst,
		otTables.ReverseChainSingleSubst,
		otTables.SinglePos,
		otTables.PairPos,
		otTables.CursivePos,
		otTables.MarkBasePos,
		otTables.MarkLigPos,
		otTables.MarkMarkPos)
def mapLookups(self, lookupMap):
	pass

# Copied and trimmed down from subset.py
@add_method(otTables.ContextSubst,
		otTables.ChainContextSubst,
		otTables.ContextPos,
		otTables.ChainContextPos)
def __merge_classify_context(self):

	class ContextHelper(object):
		def __init__(self, klass, Format):
			if klass.__name__.endswith('Subst'):
				Typ = 'Sub'
				Type = 'Subst'
			else:
				Typ = 'Pos'
				Type = 'Pos'
			if klass.__name__.startswith('Chain'):
				Chain = 'Chain'
			else:
				Chain = ''
			ChainTyp = Chain+Typ

			self.Typ = Typ
			self.Type = Type
			self.Chain = Chain
			self.ChainTyp = ChainTyp

			self.LookupRecord = Type+'LookupRecord'

			if Format == 1:
				self.Rule = ChainTyp+'Rule'
				self.RuleSet = ChainTyp+'RuleSet'
			elif Format == 2:
				self.Rule = ChainTyp+'ClassRule'
				self.RuleSet = ChainTyp+'ClassSet'

	if self.Format not in [1, 2, 3]:
		return None  # Don't shoot the messenger; let it go
	if not hasattr(self.__class__, "_merge__ContextHelpers"):
		self.__class__._merge__ContextHelpers = {}
	if self.Format not in self.__class__._merge__ContextHelpers:
		helper = ContextHelper(self.__class__, self.Format)
		self.__class__._merge__ContextHelpers[self.Format] = helper
	return self.__class__._merge__ContextHelpers[self.Format]


@add_method(otTables.ContextSubst,
		otTables.ChainContextSubst,
		otTables.ContextPos,
		otTables.ChainContextPos)
def mapLookups(self, lookupMap):
	c = self.__merge_classify_context()

	if self.Format in [1, 2]:
		for rs in getattr(self, c.RuleSet):
			if not rs: continue
			for r in getattr(rs, c.Rule):
				if not r: continue
				for ll in getattr(r, c.LookupRecord):
					if not ll: continue
					ll.LookupListIndex = lookupMap[ll.LookupListIndex]
	elif self.Format == 3:
		for ll in getattr(self, c.LookupRecord):
			if not ll: continue
			ll.LookupListIndex = lookupMap[ll.LookupListIndex]
	else:
		assert 0, "unknown format: %s" % self.Format

@add_method(otTables.ExtensionSubst,
		otTables.ExtensionPos)
def mapLookups(self, lookupMap):
	if self.Format == 1:
		self.ExtSubTable.mapLookups(lookupMap)
	else:
		assert 0, "unknown format: %s" % self.Format

@add_method(otTables.Lookup)
def mapLookups(self, lookupMap):
	for st in self.SubTable:
		if not st: continue
		st.mapLookups(lookupMap)

@add_method(otTables.LookupList)
def mapLookups(self, lookupMap):
	for l in self.Lookup:
		if not l: continue
		l.mapLookups(lookupMap)

@add_method(otTables.Lookup)
def mapMarkFilteringSets(self, markFilteringSetMap):
	if self.LookupFlag & 0x0010:
		self.MarkFilteringSet = markFilteringSetMap[self.MarkFilteringSet]

@add_method(otTables.LookupList)
def mapMarkFilteringSets(self, markFilteringSetMap):
	for l in self.Lookup:
		if not l: continue
		l.mapMarkFilteringSets(markFilteringSetMap)

@add_method(otTables.Feature)
def mapLookups(self, lookupMap):
	self.LookupListIndex = [lookupMap[i] for i in self.LookupListIndex]

@add_method(otTables.FeatureList)
def mapLookups(self, lookupMap):
	for f in self.FeatureRecord:
		if not f or not f.Feature: continue
		f.Feature.mapLookups(lookupMap)

@add_method(otTables.DefaultLangSys,
		otTables.LangSys)
def mapFeatures(self, featureMap):
	self.FeatureIndex = [featureMap[i] for i in self.FeatureIndex]
	if self.ReqFeatureIndex != 65535:
		self.ReqFeatureIndex = featureMap[self.ReqFeatureIndex]

@add_method(otTables.Script)
def mapFeatures(self, featureMap):
	if self.DefaultLangSys:
		self.DefaultLangSys.mapFeatures(featureMap)
	for l in self.LangSysRecord:
		if not l or not l.LangSys: continue
		l.LangSys.mapFeatures(featureMap)

@add_method(otTables.ScriptList)
def mapFeatures(self, featureMap):
	for s in self.ScriptRecord:
		if not s or not s.Script: continue
		s.Script.mapFeatures(featureMap)

def layoutPreMerge(font):
		# Map indices to references

		GDEF = font.get('GDEF')
		GSUB = font.get('GSUB')
		GPOS = font.get('GPOS')

		for t in [GSUB, GPOS]:
			if not t: continue

			if t.table.LookupList:
				lookupMap = {i:v for i,v in enumerate(t.table.LookupList.Lookup)}
				t.table.LookupList.mapLookups(lookupMap)
				t.table.FeatureList.mapLookups(lookupMap)

				if GDEF and GDEF.table.Version >= 0x00010002:
					markFilteringSetMap = {i:v for i,v in enumerate(GDEF.table.MarkGlyphSetsDef.Coverage)}
					t.table.LookupList.mapMarkFilteringSets(markFilteringSetMap)

			if t.table.FeatureList and t.table.ScriptList:
				featureMap = {i:v for i,v in enumerate(t.table.FeatureList.FeatureRecord)}
				t.table.ScriptList.mapFeatures(featureMap)

		# TODO FeatureParams nameIDs

def layoutPostMerge(font):
		# Map references back to indices

		GDEF = font.get('GDEF')
		GSUB = font.get('GSUB')
		GPOS = font.get('GPOS')

		for t in [GSUB, GPOS]:
			if not t: continue

			if t.table.FeatureList and t.table.ScriptList:

				# Collect unregistered (new) features.
				featureMap = GregariousIdentityDict(t.table.FeatureList.FeatureRecord)
				t.table.ScriptList.mapFeatures(featureMap)

				# Record used features.
				featureMap = AttendanceRecordingIdentityDict(t.table.FeatureList.FeatureRecord)
				t.table.ScriptList.mapFeatures(featureMap)
				usedIndices = featureMap.s

				# Remove unused features
				t.table.FeatureList.FeatureRecord = [f for i,f in enumerate(t.table.FeatureList.FeatureRecord) if i in usedIndices]

				# Map back to indices.
				featureMap = NonhashableDict(t.table.FeatureList.FeatureRecord)
				t.table.ScriptList.mapFeatures(featureMap)

				t.table.FeatureList.FeatureCount = len(t.table.FeatureList.FeatureRecord)

			if t.table.LookupList:

				# Collect unregistered (new) lookups.
				lookupMap = GregariousIdentityDict(t.table.LookupList.Lookup)
				t.table.FeatureList.mapLookups(lookupMap)
				t.table.LookupList.mapLookups(lookupMap)

				# Record used lookups.
				lookupMap = AttendanceRecordingIdentityDict(t.table.LookupList.Lookup)
				t.table.FeatureList.mapLookups(lookupMap)
				t.table.LookupList.mapLookups(lookupMap)
				usedIndices = lookupMap.s

				# Remove unused lookups
				t.table.LookupList.Lookup = [l for i,l in enumerate(t.table.LookupList.Lookup) if i in usedIndices]

				# Map back to indices.
				lookupMap = NonhashableDict(t.table.LookupList.Lookup)
				t.table.FeatureList.mapLookups(lookupMap)
				t.table.LookupList.mapLookups(lookupMap)

				t.table.LookupList.LookupCount = len(t.table.LookupList.Lookup)

				if GDEF and GDEF.table.Version >= 0x00010002:
					markFilteringSetMap = NonhashableDict(GDEF.table.MarkGlyphSetsDef.Coverage)
					t.table.LookupList.mapMarkFilteringSets(markFilteringSetMap)

		# TODO FeatureParams nameIDs



# Copyright 2021 Behdad Esfahbod. All Rights Reserved.

def is_Default_Ignorable(u):
	# http://www.unicode.org/reports/tr44/#Default_Ignorable_Code_Point
	#
	# TODO Move me to unicodedata module and autogenerate.
	#
	# Unicode 14.0:
	# $ grep '; Default_Ignorable_Code_Point ' DerivedCoreProperties.txt | sed 's/;.*#/#/'
	# 00AD          # Cf       SOFT HYPHEN
	# 034F          # Mn       COMBINING GRAPHEME JOINER
	# 061C          # Cf       ARABIC LETTER MARK
	# 115F..1160    # Lo   [2] HANGUL CHOSEONG FILLER..HANGUL JUNGSEONG FILLER
	# 17B4..17B5    # Mn   [2] KHMER VOWEL INHERENT AQ..KHMER VOWEL INHERENT AA
	# 180B..180D    # Mn   [3] MONGOLIAN FREE VARIATION SELECTOR ONE..MONGOLIAN FREE VARIATION SELECTOR THREE
	# 180E          # Cf       MONGOLIAN VOWEL SEPARATOR
	# 180F          # Mn       MONGOLIAN FREE VARIATION SELECTOR FOUR
	# 200B..200F    # Cf   [5] ZERO WIDTH SPACE..RIGHT-TO-LEFT MARK
	# 202A..202E    # Cf   [5] LEFT-TO-RIGHT EMBEDDING..RIGHT-TO-LEFT OVERRIDE
	# 2060..2064    # Cf   [5] WORD JOINER..INVISIBLE PLUS
	# 2065          # Cn       <reserved-2065>
	# 2066..206F    # Cf  [10] LEFT-TO-RIGHT ISOLATE..NOMINAL DIGIT SHAPES
	# 3164          # Lo       HANGUL FILLER
	# FE00..FE0F    # Mn  [16] VARIATION SELECTOR-1..VARIATION SELECTOR-16
	# FEFF          # Cf       ZERO WIDTH NO-BREAK SPACE
	# FFA0          # Lo       HALFWIDTH HANGUL FILLER
	# FFF0..FFF8    # Cn   [9] <reserved-FFF0>..<reserved-FFF8>
	# 1BCA0..1BCA3  # Cf   [4] SHORTHAND FORMAT LETTER OVERLAP..SHORTHAND FORMAT UP STEP
	# 1D173..1D17A  # Cf   [8] MUSICAL SYMBOL BEGIN BEAM..MUSICAL SYMBOL END PHRASE
	# E0000         # Cn       <reserved-E0000>
	# E0001         # Cf       LANGUAGE TAG
	# E0002..E001F  # Cn  [30] <reserved-E0002>..<reserved-E001F>
	# E0020..E007F  # Cf  [96] TAG SPACE..CANCEL TAG
	# E0080..E00FF  # Cn [128] <reserved-E0080>..<reserved-E00FF>
	# E0100..E01EF  # Mn [240] VARIATION SELECTOR-17..VARIATION SELECTOR-256
	# E01F0..E0FFF  # Cn [3600] <reserved-E01F0>..<reserved-E0FFF>
	return (
		u == 0x00AD or				# Cf       SOFT HYPHEN
		u == 0x034F or				# Mn       COMBINING GRAPHEME JOINER
		u == 0x061C or				# Cf       ARABIC LETTER MARK
		0x115F <= u <= 0x1160 or	# Lo   [2] HANGUL CHOSEONG FILLER..HANGUL JUNGSEONG FILLER
		0x17B4 <= u <= 0x17B5 or	# Mn   [2] KHMER VOWEL INHERENT AQ..KHMER VOWEL INHERENT AA
		0x180B <= u <= 0x180D or	# Mn   [3] MONGOLIAN FREE VARIATION SELECTOR ONE..MONGOLIAN FREE VARIATION SELECTOR THREE
		u == 0x180E or				# Cf       MONGOLIAN VOWEL SEPARATOR
		u == 0x180F or				# Mn       MONGOLIAN FREE VARIATION SELECTOR FOUR
		0x200B <= u <= 0x200F or	# Cf   [5] ZERO WIDTH SPACE..RIGHT-TO-LEFT MARK
		0x202A <= u <= 0x202E or	# Cf   [5] LEFT-TO-RIGHT EMBEDDING..RIGHT-TO-LEFT OVERRIDE
		0x2060 <= u <= 0x2064 or	# Cf   [5] WORD JOINER..INVISIBLE PLUS
		u == 0x2065 or				# Cn       <reserved-2065>
		0x2066 <= u <= 0x206F or	# Cf  [10] LEFT-TO-RIGHT ISOLATE..NOMINAL DIGIT SHAPES
		u == 0x3164 or				# Lo       HANGUL FILLER
		0xFE00 <= u <= 0xFE0F or	# Mn  [16] VARIATION SELECTOR-1..VARIATION SELECTOR-16
		u == 0xFEFF or				# Cf       ZERO WIDTH NO-BREAK SPACE
		u == 0xFFA0 or				# Lo       HALFWIDTH HANGUL FILLER
		0xFFF0 <= u <= 0xFFF8 or	# Cn   [9] <reserved-FFF0>..<reserved-FFF8>
		0x1BCA0 <= u <= 0x1BCA3 or	# Cf   [4] SHORTHAND FORMAT LETTER OVERLAP..SHORTHAND FORMAT UP STEP
		0x1D173 <= u <= 0x1D17A or	# Cf   [8] MUSICAL SYMBOL BEGIN BEAM..MUSICAL SYMBOL END PHRASE
		u == 0xE0000 or				# Cn       <reserved-E0000>
		u == 0xE0001 or				# Cf       LANGUAGE TAG
		0xE0002 <= u <= 0xE001F or	# Cn  [30] <reserved-E0002>..<reserved-E001F>
		0xE0020 <= u <= 0xE007F or	# Cf  [96] TAG SPACE..CANCEL TAG
		0xE0080 <= u <= 0xE00FF or	# Cn [128] <reserved-E0080>..<reserved-E00FF>
		0xE0100 <= u <= 0xE01EF or	# Mn [240] VARIATION SELECTOR-17..VARIATION SELECTOR-256
		0xE01F0 <= u <= 0xE0FFF or	# Cn [3600] <reserved-E01F0>..<reserved-E0FFF>
		False)



from copy import deepcopy
import string
from fontTools.colorLib.builder import LayerListBuilder, buildCOLR, buildClipList
from fontTools.misc.testTools import getXML
from fontTools.varLib.merger import COLRVariationMerger
from fontTools.varLib.models import VariationModel
from fontTools.ttLib import TTFont
from fontTools.ttLib.tables import otTables as ot
from fontTools.ttLib.tables.otBase import OTTableReader, OTTableWriter
import pytest


NO_VARIATION_INDEX = ot.NO_VARIATION_INDEX


def dump_xml(table, ttFont=None):
    xml = getXML(table.toXML, ttFont)
    print("[")
    for line in xml:
        print(f"  {line!r},")
    print("]")
    return xml


def compile_decompile(table, ttFont):
    writer = OTTableWriter(tableTag="COLR")
    # compile itself may modify a table, safer to copy it first
    table = deepcopy(table)
    table.compile(writer, ttFont)
    data = writer.getAllData()

    reader = OTTableReader(data, tableTag="COLR")
    table2 = table.__class__()
    table2.decompile(reader, ttFont)

    return table2


@pytest.fixture
def ttFont():
    font = TTFont()
    font.setGlyphOrder([".notdef"] + list(string.ascii_letters))
    return font


def build_paint(data):
    return LayerListBuilder().buildPaint(data)


class COLRVariationMergerTest:
    @pytest.mark.parametrize(
        "paints, expected_xml, expected_varIdxes",
        [
            pytest.param(
                [
                    {
                        "Format": int(ot.PaintFormat.PaintSolid),
                        "PaletteIndex": 0,
                        "Alpha": 1.0,
                    },
                    {
                        "Format": int(ot.PaintFormat.PaintSolid),
                        "PaletteIndex": 0,
                        "Alpha": 1.0,
                    },
                ],
                [
                    '<Paint Format="2"><!-- PaintSolid -->',
                    '  <PaletteIndex value="0"/>',
                    '  <Alpha value="1.0"/>',
                    "</Paint>",
                ],
                [],
                id="solid-same",
            ),
            pytest.param(
                [
                    {
                        "Format": int(ot.PaintFormat.PaintSolid),
                        "PaletteIndex": 0,
                        "Alpha": 1.0,
                    },
                    {
                        "Format": int(ot.PaintFormat.PaintSolid),
                        "PaletteIndex": 0,
                        "Alpha": 0.5,
                    },
                ],
                [
                    '<Paint Format="3"><!-- PaintVarSolid -->',
                    '  <PaletteIndex value="0"/>',
                    '  <Alpha value="1.0"/>',
                    '  <VarIndexBase value="0"/>',
                    "</Paint>",
                ],
                [0],
                id="solid-alpha",
            ),
            pytest.param(
                [
                    {
                        "Format": int(ot.PaintFormat.PaintLinearGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.PAD),
                            "ColorStop": [
                                {"StopOffset": 0.0, "PaletteIndex": 0, "Alpha": 1.0},
                                {"StopOffset": 1.0, "PaletteIndex": 1, "Alpha": 1.0},
                            ],
                        },
                        "x0": 0,
                        "y0": 0,
                        "x1": 1,
                        "y1": 1,
                        "x2": 2,
                        "y2": 2,
                    },
                    {
                        "Format": int(ot.PaintFormat.PaintLinearGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.PAD),
                            "ColorStop": [
                                {"StopOffset": 0.1, "PaletteIndex": 0, "Alpha": 1.0},
                                {"StopOffset": 0.9, "PaletteIndex": 1, "Alpha": 1.0},
                            ],
                        },
                        "x0": 0,
                        "y0": 0,
                        "x1": 1,
                        "y1": 1,
                        "x2": 2,
                        "y2": 2,
                    },
                ],
                [
                    '<Paint Format="5"><!-- PaintVarLinearGradient -->',
                    "  <ColorLine>",
                    '    <Extend value="pad"/>',
                    "    <!-- StopCount=2 -->",
                    '    <ColorStop index="0">',
                    '      <StopOffset value="0.0"/>',
                    '      <PaletteIndex value="0"/>',
                    '      <Alpha value="1.0"/>',
                    '      <VarIndexBase value="0"/>',
                    "    </ColorStop>",
                    '    <ColorStop index="1">',
                    '      <StopOffset value="1.0"/>',
                    '      <PaletteIndex value="1"/>',
                    '      <Alpha value="1.0"/>',
                    '      <VarIndexBase value="2"/>',
                    "    </ColorStop>",
                    "  </ColorLine>",
                    '  <x0 value="0"/>',
                    '  <y0 value="0"/>',
                    '  <x1 value="1"/>',
                    '  <y1 value="1"/>',
                    '  <x2 value="2"/>',
                    '  <y2 value="2"/>',
                    "  <VarIndexBase/>",
                    "</Paint>",
                ],
                [0, NO_VARIATION_INDEX, 1, NO_VARIATION_INDEX],
                id="linear_grad-stop-offsets",
            ),
            pytest.param(
                [
                    {
                        "Format": int(ot.PaintFormat.PaintLinearGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.PAD),
                            "ColorStop": [
                                {"StopOffset": 0.0, "PaletteIndex": 0, "Alpha": 1.0},
                                {"StopOffset": 1.0, "PaletteIndex": 1, "Alpha": 1.0},
                            ],
                        },
                        "x0": 0,
                        "y0": 0,
                        "x1": 1,
                        "y1": 1,
                        "x2": 2,
                        "y2": 2,
                    },
                    {
                        "Format": int(ot.PaintFormat.PaintLinearGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.PAD),
                            "ColorStop": [
                                {"StopOffset": 0.0, "PaletteIndex": 0, "Alpha": 0.5},
                                {"StopOffset": 1.0, "PaletteIndex": 1, "Alpha": 1.0},
                            ],
                        },
                        "x0": 0,
                        "y0": 0,
                        "x1": 1,
                        "y1": 1,
                        "x2": 2,
                        "y2": 2,
                    },
                ],
                [
                    '<Paint Format="5"><!-- PaintVarLinearGradient -->',
                    "  <ColorLine>",
                    '    <Extend value="pad"/>',
                    "    <!-- StopCount=2 -->",
                    '    <ColorStop index="0">',
                    '      <StopOffset value="0.0"/>',
                    '      <PaletteIndex value="0"/>',
                    '      <Alpha value="1.0"/>',
                    '      <VarIndexBase value="0"/>',
                    "    </ColorStop>",
                    '    <ColorStop index="1">',
                    '      <StopOffset value="1.0"/>',
                    '      <PaletteIndex value="1"/>',
                    '      <Alpha value="1.0"/>',
                    "      <VarIndexBase/>",
                    "    </ColorStop>",
                    "  </ColorLine>",
                    '  <x0 value="0"/>',
                    '  <y0 value="0"/>',
                    '  <x1 value="1"/>',
                    '  <y1 value="1"/>',
                    '  <x2 value="2"/>',
                    '  <y2 value="2"/>',
                    "  <VarIndexBase/>",
                    "</Paint>",
                ],
                [NO_VARIATION_INDEX, 0],
                id="linear_grad-stop[0].alpha",
            ),
            pytest.param(
                [
                    {
                        "Format": int(ot.PaintFormat.PaintLinearGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.PAD),
                            "ColorStop": [
                                {"StopOffset": 0.0, "PaletteIndex": 0, "Alpha": 1.0},
                                {"StopOffset": 1.0, "PaletteIndex": 1, "Alpha": 1.0},
                            ],
                        },
                        "x0": 0,
                        "y0": 0,
                        "x1": 1,
                        "y1": 1,
                        "x2": 2,
                        "y2": 2,
                    },
                    {
                        "Format": int(ot.PaintFormat.PaintLinearGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.PAD),
                            "ColorStop": [
                                {"StopOffset": -0.5, "PaletteIndex": 0, "Alpha": 1.0},
                                {"StopOffset": 1.0, "PaletteIndex": 1, "Alpha": 1.0},
                            ],
                        },
                        "x0": 0,
                        "y0": 0,
                        "x1": 1,
                        "y1": 1,
                        "x2": 2,
                        "y2": -200,
                    },
                ],
                [
                    '<Paint Format="5"><!-- PaintVarLinearGradient -->',
                    "  <ColorLine>",
                    '    <Extend value="pad"/>',
                    "    <!-- StopCount=2 -->",
                    '    <ColorStop index="0">',
                    '      <StopOffset value="0.0"/>',
                    '      <PaletteIndex value="0"/>',
                    '      <Alpha value="1.0"/>',
                    '      <VarIndexBase value="0"/>',
                    "    </ColorStop>",
                    '    <ColorStop index="1">',
                    '      <StopOffset value="1.0"/>',
                    '      <PaletteIndex value="1"/>',
                    '      <Alpha value="1.0"/>',
                    "      <VarIndexBase/>",
                    "    </ColorStop>",
                    "  </ColorLine>",
                    '  <x0 value="0"/>',
                    '  <y0 value="0"/>',
                    '  <x1 value="1"/>',
                    '  <y1 value="1"/>',
                    '  <x2 value="2"/>',
                    '  <y2 value="2"/>',
                    '  <VarIndexBase value="1"/>',
                    "</Paint>",
                ],
                [
                    0,
                    NO_VARIATION_INDEX,
                    NO_VARIATION_INDEX,
                    NO_VARIATION_INDEX,
                    NO_VARIATION_INDEX,
                    NO_VARIATION_INDEX,
                    1,
                ],
                id="linear_grad-stop[0].offset-y2",
            ),
            pytest.param(
                [
                    {
                        "Format": int(ot.PaintFormat.PaintRadialGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.PAD),
                            "ColorStop": [
                                {"StopOffset": 0.0, "PaletteIndex": 0, "Alpha": 1.0},
                                {"StopOffset": 1.0, "PaletteIndex": 1, "Alpha": 1.0},
                            ],
                        },
                        "x0": 0,
                        "y0": 0,
                        "r0": 0,
                        "x1": 1,
                        "y1": 1,
                        "r1": 1,
                    },
                    {
                        "Format": int(ot.PaintFormat.PaintRadialGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.PAD),
                            "ColorStop": [
                                {"StopOffset": 0.1, "PaletteIndex": 0, "Alpha": 0.6},
                                {"StopOffset": 0.9, "PaletteIndex": 1, "Alpha": 0.7},
                            ],
                        },
                        "x0": -1,
                        "y0": -2,
                        "r0": 3,
                        "x1": -4,
                        "y1": -5,
                        "r1": 6,
                    },
                ],
                [
                    '<Paint Format="7"><!-- PaintVarRadialGradient -->',
                    "  <ColorLine>",
                    '    <Extend value="pad"/>',
                    "    <!-- StopCount=2 -->",
                    '    <ColorStop index="0">',
                    '      <StopOffset value="0.0"/>',
                    '      <PaletteIndex value="0"/>',
                    '      <Alpha value="1.0"/>',
                    '      <VarIndexBase value="0"/>',
                    "    </ColorStop>",
                    '    <ColorStop index="1">',
                    '      <StopOffset value="1.0"/>',
                    '      <PaletteIndex value="1"/>',
                    '      <Alpha value="1.0"/>',
                    '      <VarIndexBase value="2"/>',
                    "    </ColorStop>",
                    "  </ColorLine>",
                    '  <x0 value="0"/>',
                    '  <y0 value="0"/>',
                    '  <r0 value="0"/>',
                    '  <x1 value="1"/>',
                    '  <y1 value="1"/>',
                    '  <r1 value="1"/>',
                    '  <VarIndexBase value="4"/>',
                    "</Paint>",
                ],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                id="radial_grad-all-different",
            ),
            pytest.param(
                [
                    {
                        "Format": int(ot.PaintFormat.PaintSweepGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.REPEAT),
                            "ColorStop": [
                                {"StopOffset": 0.4, "PaletteIndex": 0, "Alpha": 1.0},
                                {"StopOffset": 0.6, "PaletteIndex": 1, "Alpha": 1.0},
                            ],
                        },
                        "centerX": 0,
                        "centerY": 0,
                        "startAngle": 0,
                        "endAngle": 180.0,
                    },
                    {
                        "Format": int(ot.PaintFormat.PaintSweepGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.REPEAT),
                            "ColorStop": [
                                {"StopOffset": 0.4, "PaletteIndex": 0, "Alpha": 1.0},
                                {"StopOffset": 0.6, "PaletteIndex": 1, "Alpha": 1.0},
                            ],
                        },
                        "centerX": 0,
                        "centerY": 0,
                        "startAngle": 90.0,
                        "endAngle": 180.0,
                    },
                ],
                [
                    '<Paint Format="9"><!-- PaintVarSweepGradient -->',
                    "  <ColorLine>",
                    '    <Extend value="repeat"/>',
                    "    <!-- StopCount=2 -->",
                    '    <ColorStop index="0">',
                    '      <StopOffset value="0.4"/>',
                    '      <PaletteIndex value="0"/>',
                    '      <Alpha value="1.0"/>',
                    "      <VarIndexBase/>",
                    "    </ColorStop>",
                    '    <ColorStop index="1">',
                    '      <StopOffset value="0.6"/>',
                    '      <PaletteIndex value="1"/>',
                    '      <Alpha value="1.0"/>',
                    "      <VarIndexBase/>",
                    "    </ColorStop>",
                    "  </ColorLine>",
                    '  <centerX value="0"/>',
                    '  <centerY value="0"/>',
                    '  <startAngle value="0.0"/>',
                    '  <endAngle value="180.0"/>',
                    '  <VarIndexBase value="0"/>',
                    "</Paint>",
                ],
                [NO_VARIATION_INDEX, NO_VARIATION_INDEX, 0, NO_VARIATION_INDEX],
                id="sweep_grad-startAngle",
            ),
            pytest.param(
                [
                    {
                        "Format": int(ot.PaintFormat.PaintSweepGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.PAD),
                            "ColorStop": [
                                {"StopOffset": 0.0, "PaletteIndex": 0, "Alpha": 1.0},
                                {"StopOffset": 1.0, "PaletteIndex": 1, "Alpha": 1.0},
                            ],
                        },
                        "centerX": 0,
                        "centerY": 0,
                        "startAngle": 0.0,
                        "endAngle": 180.0,
                    },
                    {
                        "Format": int(ot.PaintFormat.PaintSweepGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.PAD),
                            "ColorStop": [
                                {"StopOffset": 0.0, "PaletteIndex": 0, "Alpha": 0.5},
                                {"StopOffset": 1.0, "PaletteIndex": 1, "Alpha": 0.5},
                            ],
                        },
                        "centerX": 0,
                        "centerY": 0,
                        "startAngle": 0.0,
                        "endAngle": 180.0,
                    },
                ],
                [
                    '<Paint Format="9"><!-- PaintVarSweepGradient -->',
                    "  <ColorLine>",
                    '    <Extend value="pad"/>',
                    "    <!-- StopCount=2 -->",
                    '    <ColorStop index="0">',
                    '      <StopOffset value="0.0"/>',
                    '      <PaletteIndex value="0"/>',
                    '      <Alpha value="1.0"/>',
                    '      <VarIndexBase value="0"/>',
                    "    </ColorStop>",
                    '    <ColorStop index="1">',
                    '      <StopOffset value="1.0"/>',
                    '      <PaletteIndex value="1"/>',
                    '      <Alpha value="1.0"/>',
                    '      <VarIndexBase value="0"/>',
                    "    </ColorStop>",
                    "  </ColorLine>",
                    '  <centerX value="0"/>',
                    '  <centerY value="0"/>',
                    '  <startAngle value="0.0"/>',
                    '  <endAngle value="180.0"/>',
                    "  <VarIndexBase/>",
                    "</Paint>",
                ],
                [NO_VARIATION_INDEX, 0],
                id="sweep_grad-stops-alpha-reuse-varidxbase",
            ),
            pytest.param(
                [
                    {
                        "Format": int(ot.PaintFormat.PaintTransform),
                        "Paint": {
                            "Format": int(ot.PaintFormat.PaintRadialGradient),
                            "ColorLine": {
                                "Extend": int(ot.ExtendMode.PAD),
                                "ColorStop": [
                                    {
                                        "StopOffset": 0.0,
                                        "PaletteIndex": 0,
                                        "Alpha": 1.0,
                                    },
                                    {
                                        "StopOffset": 1.0,
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                ],
                            },
                            "x0": 0,
                            "y0": 0,
                            "r0": 0,
                            "x1": 1,
                            "y1": 1,
                            "r1": 1,
                        },
                        "Transform": {
                            "xx": 1.0,
                            "xy": 0.0,
                            "yx": 0.0,
                            "yy": 1.0,
                            "dx": 0.0,
                            "dy": 0.0,
                        },
                    },
                    {
                        "Format": int(ot.PaintFormat.PaintTransform),
                        "Paint": {
                            "Format": int(ot.PaintFormat.PaintRadialGradient),
                            "ColorLine": {
                                "Extend": int(ot.ExtendMode.PAD),
                                "ColorStop": [
                                    {
                                        "StopOffset": 0.0,
                                        "PaletteIndex": 0,
                                        "Alpha": 1.0,
                                    },
                                    {
                                        "StopOffset": 1.0,
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                ],
                            },
                            "x0": 0,
                            "y0": 0,
                            "r0": 0,
                            "x1": 1,
                            "y1": 1,
                            "r1": 1,
                        },
                        "Transform": {
                            "xx": 1.0,
                            "xy": 0.0,
                            "yx": 0.0,
                            "yy": 0.5,
                            "dx": 0.0,
                            "dy": -100.0,
                        },
                    },
                ],
                [
                    '<Paint Format="13"><!-- PaintVarTransform -->',
                    '  <Paint Format="6"><!-- PaintRadialGradient -->',
                    "    <ColorLine>",
                    '      <Extend value="pad"/>',
                    "      <!-- StopCount=2 -->",
                    '      <ColorStop index="0">',
                    '        <StopOffset value="0.0"/>',
                    '        <PaletteIndex value="0"/>',
                    '        <Alpha value="1.0"/>',
                    "      </ColorStop>",
                    '      <ColorStop index="1">',
                    '        <StopOffset value="1.0"/>',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    "      </ColorStop>",
                    "    </ColorLine>",
                    '    <x0 value="0"/>',
                    '    <y0 value="0"/>',
                    '    <r0 value="0"/>',
                    '    <x1 value="1"/>',
                    '    <y1 value="1"/>',
                    '    <r1 value="1"/>',
                    "  </Paint>",
                    "  <Transform>",
                    '    <xx value="1.0"/>',
                    '    <yx value="0.0"/>',
                    '    <xy value="0.0"/>',
                    '    <yy value="1.0"/>',
                    '    <dx value="0.0"/>',
                    '    <dy value="0.0"/>',
                    '    <VarIndexBase value="0"/>',
                    "  </Transform>",
                    "</Paint>",
                ],
                [
                    NO_VARIATION_INDEX,
                    NO_VARIATION_INDEX,
                    NO_VARIATION_INDEX,
                    0,
                    NO_VARIATION_INDEX,
                    1,
                ],
                id="transform-yy-dy",
            ),
            pytest.param(
                [
                    {
                        "Format": ot.PaintFormat.PaintTransform,
                        "Paint": {
                            "Format": ot.PaintFormat.PaintSweepGradient,
                            "ColorLine": {
                                "Extend": ot.ExtendMode.PAD,
                                "ColorStop": [
                                    {"StopOffset": 0.0, "PaletteIndex": 0},
                                    {
                                        "StopOffset": 1.0,
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                ],
                            },
                            "centerX": 0,
                            "centerY": 0,
                            "startAngle": 0,
                            "endAngle": 360,
                        },
                        "Transform": (1.0, 0, 0, 1.0, 0, 0),
                    },
                    {
                        "Format": ot.PaintFormat.PaintTransform,
                        "Paint": {
                            "Format": ot.PaintFormat.PaintSweepGradient,
                            "ColorLine": {
                                "Extend": ot.ExtendMode.PAD,
                                "ColorStop": [
                                    {"StopOffset": 0.0, "PaletteIndex": 0},
                                    {
                                        "StopOffset": 1.0,
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                ],
                            },
                            "centerX": 256,
                            "centerY": 0,
                            "startAngle": 0,
                            "endAngle": 360,
                        },
                        # Transform.xx below produces the same VarStore delta as the
                        # above PaintSweepGradient's centerX because, when Fixed16.16
                        # is converted to integer, it becomes:
                        # floatToFixed(1.00390625, 16) == 256
                        # Because there is overlap between the varIdxes of the
                        # PaintVarTransform's Affine2x3 and the PaintSweepGradient's
                        # the VarIndexBase is reused (0 for both)
                        "Transform": (1.00390625, 0, 0, 1.0, 10, 0),
                    },
                ],
                [
                    '<Paint Format="13"><!-- PaintVarTransform -->',
                    '  <Paint Format="9"><!-- PaintVarSweepGradient -->',
                    "    <ColorLine>",
                    '      <Extend value="pad"/>',
                    "      <!-- StopCount=2 -->",
                    '      <ColorStop index="0">',
                    '        <StopOffset value="0.0"/>',
                    '        <PaletteIndex value="0"/>',
                    '        <Alpha value="1.0"/>',
                    "        <VarIndexBase/>",
                    "      </ColorStop>",
                    '      <ColorStop index="1">',
                    '        <StopOffset value="1.0"/>',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    "        <VarIndexBase/>",
                    "      </ColorStop>",
                    "    </ColorLine>",
                    '    <centerX value="0"/>',
                    '    <centerY value="0"/>',
                    '    <startAngle value="0.0"/>',
                    '    <endAngle value="360.0"/>',
                    '    <VarIndexBase value="0"/>',
                    "  </Paint>",
                    "  <Transform>",
                    '    <xx value="1.0"/>',
                    '    <yx value="0.0"/>',
                    '    <xy value="0.0"/>',
                    '    <yy value="1.0"/>',
                    '    <dx value="0.0"/>',
                    '    <dy value="0.0"/>',
                    '    <VarIndexBase value="0"/>',
                    "  </Transform>",
                    "</Paint>",
                ],
                [
                    0,
                    NO_VARIATION_INDEX,
                    NO_VARIATION_INDEX,
                    NO_VARIATION_INDEX,
                    1,
                    NO_VARIATION_INDEX,
                ],
                id="transform-xx-sweep_grad-centerx-same-varidxbase",
            ),
        ],
    )
    def test_merge_Paint(self, paints, ttFont, expected_xml, expected_varIdxes):
        paints = [build_paint(p) for p in paints]
        out = deepcopy(paints[0])

        model = VariationModel([{}, {"ZZZZ": 1.0}])
        merger = COLRVariationMerger(model, ["ZZZZ"], ttFont)

        merger.mergeThings(out, paints)

        assert compile_decompile(out, ttFont) == out
        assert dump_xml(out, ttFont) == expected_xml
        assert merger.varIdxes == expected_varIdxes

    def test_merge_ClipList(self, ttFont):
        clipLists = [
            buildClipList(clips)
            for clips in [
                {
                    "A": (0, 0, 1000, 1000),
                    "B": (0, 0, 1000, 1000),
                    "C": (0, 0, 1000, 1000),
                    "D": (0, 0, 1000, 1000),
                },
                {
                    # non-default masters' clip boxes can be 'sparse'
                    # (i.e. can omit explicit clip box for some glyphs)
                    # "A": (0, 0, 1000, 1000),
                    "B": (10, 0, 1000, 1000),
                    "C": (20, 20, 1020, 1020),
                    "D": (20, 20, 1020, 1020),
                },
            ]
        ]
        out = deepcopy(clipLists[0])

        model = VariationModel([{}, {"ZZZZ": 1.0}])
        merger = COLRVariationMerger(model, ["ZZZZ"], ttFont)

        merger.mergeThings(out, clipLists)

        assert compile_decompile(out, ttFont) == out
        assert dump_xml(out, ttFont) == [
            '<ClipList Format="1">',
            "  <Clip>",
            '    <Glyph value="A"/>',
            '    <ClipBox Format="1">',
            '      <xMin value="0"/>',
            '      <yMin value="0"/>',
            '      <xMax value="1000"/>',
            '      <yMax value="1000"/>',
            "    </ClipBox>",
            "  </Clip>",
            "  <Clip>",
            '    <Glyph value="B"/>',
            '    <ClipBox Format="2">',
            '      <xMin value="0"/>',
            '      <yMin value="0"/>',
            '      <xMax value="1000"/>',
            '      <yMax value="1000"/>',
            '      <VarIndexBase value="0"/>',
            "    </ClipBox>",
            "  </Clip>",
            "  <Clip>",
            '    <Glyph value="C"/>',
            '    <Glyph value="D"/>',
            '    <ClipBox Format="2">',
            '      <xMin value="0"/>',
            '      <yMin value="0"/>',
            '      <xMax value="1000"/>',
            '      <yMax value="1000"/>',
            '      <VarIndexBase value="4"/>',
            "    </ClipBox>",
            "  </Clip>",
            "</ClipList>",
        ]
        assert merger.varIdxes == [
            0,
            NO_VARIATION_INDEX,
            NO_VARIATION_INDEX,
            NO_VARIATION_INDEX,
            1,
            1,
            1,
            1,
        ]

    @pytest.mark.parametrize(
        "master_layer_reuse",
        [
            pytest.param(False, id="no-reuse"),
            pytest.param(True, id="with-reuse"),
        ],
    )
    @pytest.mark.parametrize(
        "color_glyphs, output_layer_reuse, expected_xml, expected_varIdxes",
        [
            pytest.param(
                [
                    {
                        "A": {
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 0,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                    },
                    {
                        "A": {
                            "Format": ot.PaintFormat.PaintColrLayers,
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 0,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                    },
                ],
                False,
                [
                    "<COLR>",
                    '  <Version value="1"/>',
                    "  <!-- BaseGlyphRecordCount=0 -->",
                    "  <!-- LayerRecordCount=0 -->",
                    "  <BaseGlyphList>",
                    "    <!-- BaseGlyphCount=1 -->",
                    '    <BaseGlyphPaintRecord index="0">',
                    '      <BaseGlyph value="A"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="2"/>',
                    '        <FirstLayerIndex value="0"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    "  </BaseGlyphList>",
                    "  <LayerList>",
                    "    <!-- LayerCount=2 -->",
                    '    <Paint index="0" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="0"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="1" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    "  </LayerList>",
                    "</COLR>",
                ],
                [],
                id="no-variation",
            ),
            pytest.param(
                [
                    {
                        "A": {
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 0,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                        "C": {
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 3,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                    },
                    {
                        # NOTE: 'A' is missing from non-default master
                        "C": {
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 0.5,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 3,
                                        "Alpha": 0.5,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                    },
                ],
                False,
                [
                    "<COLR>",
                    '  <Version value="1"/>',
                    "  <!-- BaseGlyphRecordCount=0 -->",
                    "  <!-- LayerRecordCount=0 -->",
                    "  <BaseGlyphList>",
                    "    <!-- BaseGlyphCount=2 -->",
                    '    <BaseGlyphPaintRecord index="0">',
                    '      <BaseGlyph value="A"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="2"/>',
                    '        <FirstLayerIndex value="0"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    '    <BaseGlyphPaintRecord index="1">',
                    '      <BaseGlyph value="C"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="2"/>',
                    '        <FirstLayerIndex value="2"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    "  </BaseGlyphList>",
                    "  <LayerList>",
                    "    <!-- LayerCount=4 -->",
                    '    <Paint index="0" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="0"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="1" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="2" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="3"><!-- PaintVarSolid -->',
                    '        <PaletteIndex value="2"/>',
                    '        <Alpha value="1.0"/>',
                    '        <VarIndexBase value="0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="3" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="3"><!-- PaintVarSolid -->',
                    '        <PaletteIndex value="3"/>',
                    '        <Alpha value="1.0"/>',
                    '        <VarIndexBase value="0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    "  </LayerList>",
                    "</COLR>",
                ],
                [0],
                id="sparse-masters",
            ),
            pytest.param(
                [
                    {
                        "A": {
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 0,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                        "C": {
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                # 'C' reuses layers 1-3 from 'A'
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                        "D": {  # identical to 'C'
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                        "E": {  # superset of 'C' or 'D'
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 3,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                    },
                    {
                        # NOTE: 'A' is missing from non-default master
                        "C": {
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 0.5,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 0.5,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                        "D": {  # same as 'C'
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 0.5,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 0.5,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                        "E": {  # first two layers vary the same way as 'C' or 'D'
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 0.5,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 0.5,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 3,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                    },
                ],
                True,  # reuse
                [
                    "<COLR>",
                    '  <Version value="1"/>',
                    "  <!-- BaseGlyphRecordCount=0 -->",
                    "  <!-- LayerRecordCount=0 -->",
                    "  <BaseGlyphList>",
                    "    <!-- BaseGlyphCount=4 -->",
                    '    <BaseGlyphPaintRecord index="0">',
                    '      <BaseGlyph value="A"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="3"/>',
                    '        <FirstLayerIndex value="0"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    '    <BaseGlyphPaintRecord index="1">',
                    '      <BaseGlyph value="C"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="2"/>',
                    '        <FirstLayerIndex value="3"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    '    <BaseGlyphPaintRecord index="2">',
                    '      <BaseGlyph value="D"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="2"/>',
                    '        <FirstLayerIndex value="3"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    '    <BaseGlyphPaintRecord index="3">',
                    '      <BaseGlyph value="E"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="2"/>',
                    '        <FirstLayerIndex value="5"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    "  </BaseGlyphList>",
                    "  <LayerList>",
                    "    <!-- LayerCount=7 -->",
                    '    <Paint index="0" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="0"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="1" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="2" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="2"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="3" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="3"><!-- PaintVarSolid -->',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    '        <VarIndexBase value="0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="4" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="3"><!-- PaintVarSolid -->',
                    '        <PaletteIndex value="2"/>',
                    '        <Alpha value="1.0"/>',
                    '        <VarIndexBase value="0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="5" Format="1"><!-- PaintColrLayers -->',
                    '      <NumLayers value="2"/>',
                    '      <FirstLayerIndex value="3"/>',
                    "    </Paint>",
                    '    <Paint index="6" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="3"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    "  </LayerList>",
                    "</COLR>",
                ],
                [0],
                id="sparse-masters-with-reuse",
            ),
            pytest.param(
                [
                    {
                        "A": {
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 0,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                        "C": {  # 'C' shares layer 1 and 2 with 'A'
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                    },
                    {
                        "A": {
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 0,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 0.9,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                        "C": {
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 0.5,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                    },
                ],
                True,
                [
                    # a different Alpha variation is applied to a shared layer between
                    # 'A' and 'C' and thus they are no longer shared.
                    "<COLR>",
                    '  <Version value="1"/>',
                    "  <!-- BaseGlyphRecordCount=0 -->",
                    "  <!-- LayerRecordCount=0 -->",
                    "  <BaseGlyphList>",
                    "    <!-- BaseGlyphCount=2 -->",
                    '    <BaseGlyphPaintRecord index="0">',
                    '      <BaseGlyph value="A"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="3"/>',
                    '        <FirstLayerIndex value="0"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    '    <BaseGlyphPaintRecord index="1">',
                    '      <BaseGlyph value="C"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="2"/>',
                    '        <FirstLayerIndex value="3"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    "  </BaseGlyphList>",
                    "  <LayerList>",
                    "    <!-- LayerCount=5 -->",
                    '    <Paint index="0" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="0"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="1" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="3"><!-- PaintVarSolid -->',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    '        <VarIndexBase value="0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="2" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="2"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="3" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="3"><!-- PaintVarSolid -->',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    '        <VarIndexBase value="1"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="4" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="2"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    "  </LayerList>",
                    "</COLR>",
                ],
                [0, 1],
                id="shared-master-layers-different-variations",
            ),
        ],
    )
    def test_merge_full_table(
        self,
        color_glyphs,
        ttFont,
        expected_xml,
        expected_varIdxes,
        master_layer_reuse,
        output_layer_reuse,
    ):
        master_ttfs = [deepcopy(ttFont) for _ in range(len(color_glyphs))]
        for ttf, glyphs in zip(master_ttfs, color_glyphs):
            # merge algorithm is expected to work the same even if the master COLRs
            # may differ as to the layer reuse, hence we try both ways
            ttf["COLR"] = buildCOLR(glyphs, allowLayerReuse=master_layer_reuse)
        vf = deepcopy(master_ttfs[0])

        model = VariationModel([{}, {"ZZZZ": 1.0}])
        merger = COLRVariationMerger(
            model, ["ZZZZ"], vf, allowLayerReuse=output_layer_reuse
        )

        merger.mergeTables(vf, master_ttfs)

        out = vf["COLR"].table

        assert compile_decompile(out, vf) == out
        assert dump_xml(out, vf) == expected_xml
        assert merger.varIdxes == expected_varIdxes

    @pytest.mark.parametrize(
        "color_glyphs, before_xml, expected_xml",
        [
            pytest.param(
                {
                    "A": {
                        "Format": int(ot.PaintFormat.PaintColrLayers),
                        "Layers": [
                            {
                                "Format": int(ot.PaintFormat.PaintGlyph),
                                "Paint": {
                                    "Format": int(ot.PaintFormat.PaintSolid),
                                    "PaletteIndex": 0,
                                    "Alpha": 1.0,
                                },
                                "Glyph": "B",
                            },
                            {
                                "Format": int(ot.PaintFormat.PaintGlyph),
                                "Paint": {
                                    "Format": int(ot.PaintFormat.PaintSolid),
                                    "PaletteIndex": 1,
                                    "Alpha": 1.0,
                                },
                                "Glyph": "C",
                            },
                            {
                                "Format": int(ot.PaintFormat.PaintGlyph),
                                "Paint": {
                                    "Format": int(ot.PaintFormat.PaintSolid),
                                    "PaletteIndex": 2,
                                    "Alpha": 1.0,
                                },
                                "Glyph": "D",
                            },
                        ],
                    },
                    "E": {
                        "Format": int(ot.PaintFormat.PaintColrLayers),
                        "Layers": [
                            {
                                "Format": int(ot.PaintFormat.PaintGlyph),
                                "Paint": {
                                    "Format": int(ot.PaintFormat.PaintSolid),
                                    "PaletteIndex": 1,
                                    "Alpha": 1.0,
                                },
                                "Glyph": "C",
                            },
                            {
                                "Format": int(ot.PaintFormat.PaintGlyph),
                                "Paint": {
                                    "Format": int(ot.PaintFormat.PaintSolid),
                                    "PaletteIndex": 2,
                                    "Alpha": 1.0,
                                },
                                "Glyph": "D",
                            },
                            {
                                "Format": int(ot.PaintFormat.PaintGlyph),
                                "Paint": {
                                    "Format": int(ot.PaintFormat.PaintSolid),
                                    "PaletteIndex": 3,
                                    "Alpha": 1.0,
                                },
                                "Glyph": "F",
                            },
                        ],
                    },
                    "G": {
                        "Format": int(ot.PaintFormat.PaintColrGlyph),
                        "Glyph": "E",
                    },
                },
                [
                    "<COLR>",
                    '  <Version value="1"/>',
                    "  <!-- BaseGlyphRecordCount=0 -->",
                    "  <!-- LayerRecordCount=0 -->",
                    "  <BaseGlyphList>",
                    "    <!-- BaseGlyphCount=3 -->",
                    '    <BaseGlyphPaintRecord index="0">',
                    '      <BaseGlyph value="A"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="3"/>',
                    '        <FirstLayerIndex value="0"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    '    <BaseGlyphPaintRecord index="1">',
                    '      <BaseGlyph value="E"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="2"/>',
                    '        <FirstLayerIndex value="3"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    '    <BaseGlyphPaintRecord index="2">',
                    '      <BaseGlyph value="G"/>',
                    '      <Paint Format="11"><!-- PaintColrGlyph -->',
                    '        <Glyph value="E"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    "  </BaseGlyphList>",
                    "  <LayerList>",
                    "    <!-- LayerCount=5 -->",
                    '    <Paint index="0" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="0"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="1" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="C"/>',
                    "    </Paint>",
                    '    <Paint index="2" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="2"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="D"/>',
                    "    </Paint>",
                    '    <Paint index="3" Format="1"><!-- PaintColrLayers -->',
                    '      <NumLayers value="2"/>',
                    '      <FirstLayerIndex value="1"/>',
                    "    </Paint>",
                    '    <Paint index="4" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="3"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="F"/>',
                    "    </Paint>",
                    "  </LayerList>",
                    "</COLR>",
                ],
                [
                    "<COLR>",
                    '  <Version value="1"/>',
                    "  <!-- BaseGlyphRecordCount=0 -->",
                    "  <!-- LayerRecordCount=0 -->",
                    "  <BaseGlyphList>",
                    "    <!-- BaseGlyphCount=3 -->",
                    '    <BaseGlyphPaintRecord index="0">',
                    '      <BaseGlyph value="A"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="3"/>',
                    '        <FirstLayerIndex value="0"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    '    <BaseGlyphPaintRecord index="1">',
                    '      <BaseGlyph value="E"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="3"/>',
                    '        <FirstLayerIndex value="3"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    '    <BaseGlyphPaintRecord index="2">',
                    '      <BaseGlyph value="G"/>',
                    '      <Paint Format="11"><!-- PaintColrGlyph -->',
                    '        <Glyph value="E"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    "  </BaseGlyphList>",
                    "  <LayerList>",
                    "    <!-- LayerCount=6 -->",
                    '    <Paint index="0" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="0"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="1" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="C"/>',
                    "    </Paint>",
                    '    <Paint index="2" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="2"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="D"/>',
                    "    </Paint>",
                    '    <Paint index="3" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="C"/>',
                    "    </Paint>",
                    '    <Paint index="4" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="2"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="D"/>',
                    "    </Paint>",
                    '    <Paint index="5" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="3"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="F"/>',
                    "    </Paint>",
                    "  </LayerList>",
                    "</COLR>",
                ],
                id="simple-reuse",
            ),
            pytest.param(
                {
                    "A": {
                        "Format": int(ot.PaintFormat.PaintGlyph),
                        "Paint": {
                            "Format": int(ot.PaintFormat.PaintSolid),
                            "PaletteIndex": 0,
                            "Alpha": 1.0,
                        },
                        "Glyph": "B",
                    },
                },
                [
                    "<COLR>",
                    '  <Version value="1"/>',
                    "  <!-- BaseGlyphRecordCount=0 -->",
                    "  <!-- LayerRecordCount=0 -->",
                    "  <BaseGlyphList>",
                    "    <!-- BaseGlyphCount=1 -->",
                    '    <BaseGlyphPaintRecord index="0">',
                    '      <BaseGlyph value="A"/>',
                    '      <Paint Format="10"><!-- PaintGlyph -->',
                    '        <Paint Format="2"><!-- PaintSolid -->',
                    '          <PaletteIndex value="0"/>',
                    '          <Alpha value="1.0"/>',
                    "        </Paint>",
                    '        <Glyph value="B"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    "  </BaseGlyphList>",
                    "</COLR>",
                ],
                [
                    "<COLR>",
                    '  <Version value="1"/>',
                    "  <!-- BaseGlyphRecordCount=0 -->",
                    "  <!-- LayerRecordCount=0 -->",
                    "  <BaseGlyphList>",
                    "    <!-- BaseGlyphCount=1 -->",
                    '    <BaseGlyphPaintRecord index="0">',
                    '      <BaseGlyph value="A"/>',
                    '      <Paint Format="10"><!-- PaintGlyph -->',
                    '        <Paint Format="2"><!-- PaintSolid -->',
                    '          <PaletteIndex value="0"/>',
                    '          <Alpha value="1.0"/>',
                    "        </Paint>",
                    '        <Glyph value="B"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    "  </BaseGlyphList>",
                    "</COLR>",
                ],
                id="no-layer-list",
            ),
        ],
    )
    def test_expandPaintColrLayers(
        self, color_glyphs, ttFont, before_xml, expected_xml
    ):
        colr = buildCOLR(color_glyphs, allowLayerReuse=True)

        assert dump_xml(colr.table, ttFont) == before_xml

        before_layer_count = 0
        reuses_colr_layers = False
        if colr.table.LayerList:
            before_layer_count = len(colr.table.LayerList.Paint)
            reuses_colr_layers = any(
                p.Format == ot.PaintFormat.PaintColrLayers
                for p in colr.table.LayerList.Paint
            )

        COLRVariationMerger.expandPaintColrLayers(colr.table)

        assert dump_xml(colr.table, ttFont) == expected_xml

        after_layer_count = (
            0 if not colr.table.LayerList else len(colr.table.LayerList.Paint)
        )

        if reuses_colr_layers:
            assert not any(
                p.Format == ot.PaintFormat.PaintColrLayers
                for p in colr.table.LayerList.Paint
            )
            assert after_layer_count > before_layer_count
        else:
            assert after_layer_count == before_layer_count

        if colr.table.LayerList:
            assert len({id(p) for p in colr.table.LayerList.Paint}) == after_layer_count



# Copyright 2013 Google, Inc. All Rights Reserved.
#
# Google Author(s): Behdad Esfahbod, Roozbeh Pournader


class Options(object):

	class UnknownOptionError(Exception):
		pass

	def __init__(self, **kwargs):

		self.verbose = False
		self.timing = False
		self.drop_tables = []

		self.set(**kwargs)

	def set(self, **kwargs):
		for k,v in kwargs.items():
			if not hasattr(self, k):
				raise self.UnknownOptionError("Unknown option '%s'" % k)
			setattr(self, k, v)

	def parse_opts(self, argv, ignore_unknown=[]):
		ret = []
		opts = {}
		for a in argv:
			orig_a = a
			if not a.startswith('--'):
				ret.append(a)
				continue
			a = a[2:]
			i = a.find('=')
			op = '='
			if i == -1:
				if a.startswith("no-"):
					k = a[3:]
					v = False
				else:
					k = a
					v = True
			else:
				k = a[:i]
				if k[-1] in "-+":
					op = k[-1]+'='  # Ops is '-=' or '+=' now.
					k = k[:-1]
				v = a[i+1:]
			ok = k
			k = k.replace('-', '_')
			if not hasattr(self, k):
				if ignore_unknown is True or ok in ignore_unknown:
					ret.append(orig_a)
					continue
				else:
					raise self.UnknownOptionError("Unknown option '%s'" % a)

			ov = getattr(self, k)
			if isinstance(ov, bool):
				v = bool(v)
			elif isinstance(ov, int):
				v = int(v)
			elif isinstance(ov, list):
				vv = v.split(',')
				if vv == ['']:
					vv = []
				vv = [int(x, 0) if len(x) and x[0] in "0123456789" else x for x in vv]
				if op == '=':
					v = vv
				elif op == '+=':
					v = ov
					v.extend(vv)
				elif op == '-=':
					v = ov
					for x in vv:
						if x in v:
							v.remove(x)
				else:
					assert 0

			opts[k] = v
		self.set(**opts)

		return ret




import io
import itertools
from fontTools import ttLib
from fontTools.ttLib.tables._g_l_y_f import Glyph
from fontTools.fontBuilder import FontBuilder
from fontTools.merge import Merger, main as merge_main
import difflib
import os
import re
import shutil
import sys
import tempfile
import unittest
import pathlib
import pytest


class MergeIntegrationTest(unittest.TestCase):
	def setUp(self):
		self.tempdir = None
		self.num_tempfiles = 0

	def tearDown(self):
		if self.tempdir:
			shutil.rmtree(self.tempdir)

	@staticmethod
	def getpath(testfile):
		path, _ = os.path.split(__file__)
		return os.path.join(path, "data", testfile)

	def temp_path(self, suffix):
		if not self.tempdir:
			self.tempdir = tempfile.mkdtemp()
		self.num_tempfiles += 1
		return os.path.join(self.tempdir, "tmp%d%s" % (self.num_tempfiles, suffix))

	IGNORED_LINES_RE = re.compile(
		"^(<ttFont |    <(checkSumAdjustment|created|modified) ).*"
	)
	def read_ttx(self, path):
		lines = []
		with open(path, "r", encoding="utf-8") as ttx:
			for line in ttx.readlines():
				# Elide lines with data that often change.
				if self.IGNORED_LINES_RE.match(line):
					lines.append("\n")
				else:
					lines.append(line.rstrip() + "\n")
		return lines

	def expect_ttx(self, font, expected_ttx, tables=None):
		path = self.temp_path(suffix=".ttx")
		font.saveXML(path, tables=tables)
		actual = self.read_ttx(path)
		expected = self.read_ttx(expected_ttx)
		if actual != expected:
			for line in difflib.unified_diff(
					expected, actual, fromfile=expected_ttx, tofile=path):
				sys.stdout.write(line)
			self.fail("TTX output is different from expected")

	def compile_font(self, path, suffix):
		savepath = self.temp_path(suffix=suffix)
		font = ttLib.TTFont(recalcBBoxes=False, recalcTimestamp=False)
		font.importXML(path)
		font.save(savepath, reorderTables=None)
		return font, savepath

# -----
# Tests
# -----

	def test_merge_cff(self):
		_, fontpath1 = self.compile_font(self.getpath("CFFFont1.ttx"), ".otf")
		_, fontpath2 = self.compile_font(self.getpath("CFFFont2.ttx"), ".otf")
		mergedpath = self.temp_path(".otf")
		merge_main([fontpath1, fontpath2, "--output-file=%s" % mergedpath])
		mergedfont = ttLib.TTFont(mergedpath)
		self.expect_ttx(mergedfont, self.getpath("CFFFont_expected.ttx"))


class gaspMergeUnitTest(unittest.TestCase):
	def setUp(self):
		self.merger = Merger()

		self.table1 = ttLib.newTable('gasp')
		self.table1.version = 1
		self.table1.gaspRange = {
			0x8: 0xA ,
			0x10: 0x5,
		}

		self.table2 = ttLib.newTable('gasp')
		self.table2.version = 1
		self.table2.gaspRange = {
			0x6: 0xB ,
			0xFF: 0x4,
		}

		self.result = ttLib.newTable('gasp')

	def test_gasp_merge_basic(self):
		result = self.result.merge(self.merger, [self.table1, self.table2])
		self.assertEqual(result, self.table1)

		result = self.result.merge(self.merger, [self.table2, self.table1])
		self.assertEqual(result, self.table2)

	def test_gasp_merge_notImplemented(self):
		result = self.result.merge(self.merger, [NotImplemented, self.table1])
		self.assertEqual(result, NotImplemented)

		result = self.result.merge(self.merger, [self.table1, NotImplemented])
		self.assertEqual(result, self.table1)


class CmapMergeUnitTest(unittest.TestCase):
	def setUp(self):
		self.merger = Merger()
		self.table1 = ttLib.newTable('cmap')
		self.table2 = ttLib.newTable('cmap')
		self.mergedTable = ttLib.newTable('cmap')
		pass

	def tearDown(self):
		pass


	def makeSubtable(self, format, platformID, platEncID, cmap):
		module = ttLib.getTableModule('cmap')
		subtable = module.cmap_classes[format](format)
		(subtable.platformID,
			subtable.platEncID,
			subtable.language,
			subtable.cmap) = (platformID, platEncID, 0, cmap)
		return subtable

	# 4-3-1 table merged with 12-3-10 table with no dupes with codepoints outside BMP
	def test_cmap_merge_no_dupes(self):
		table1 = self.table1
		table2 = self.table2
		mergedTable = self.mergedTable

		cmap1 = {0x2603: 'SNOWMAN'}
		table1.tables = [self.makeSubtable(4,3,1, cmap1)]

		cmap2 = {0x26C4: 'SNOWMAN WITHOUT SNOW'}
		cmap2Extended = {0x1F93C: 'WRESTLERS'}
		cmap2Extended.update(cmap2)
		table2.tables = [self.makeSubtable(4,3,1, cmap2), self.makeSubtable(12,3,10, cmap2Extended)]

		self.merger.alternateGlyphsPerFont = [{},{}]
		mergedTable.merge(self.merger, [table1, table2])

		expectedCmap = cmap2.copy()
		expectedCmap.update(cmap1)
		expectedCmapExtended = cmap2Extended.copy()
		expectedCmapExtended.update(cmap1)
		self.assertEqual(mergedTable.numSubTables, 2)
		self.assertEqual([(table.format, table.platformID, table.platEncID, table.language) for table in mergedTable.tables],
			[(4,3,1,0),(12,3,10,0)])
		self.assertEqual(mergedTable.tables[0].cmap, expectedCmap)
		self.assertEqual(mergedTable.tables[1].cmap, expectedCmapExtended)

	# Tests Issue #322
	def test_cmap_merge_three_dupes(self):
		table1 = self.table1
		table2 = self.table2
		mergedTable = self.mergedTable

		cmap1 = {0x20: 'space#0', 0xA0: 'space#0'}
		table1.tables = [self.makeSubtable(4,3,1,cmap1)]
		cmap2 = {0x20: 'space#1', 0xA0: 'uni00A0#1'}
		table2.tables = [self.makeSubtable(4,3,1,cmap2)]

		self.merger.duplicateGlyphsPerFont = [{},{}]
		mergedTable.merge(self.merger, [table1, table2])

		expectedCmap = cmap1.copy()
		self.assertEqual(mergedTable.numSubTables, 1)
		table = mergedTable.tables[0]
		self.assertEqual((table.format, table.platformID, table.platEncID, table.language), (4,3,1,0))
		self.assertEqual(table.cmap, expectedCmap)
		self.assertEqual(self.merger.duplicateGlyphsPerFont, [{}, {'space#0': 'space#1'}])


def _compile(ttFont):
	buf = io.BytesIO()
	ttFont.save(buf)
	buf.seek(0)
	return buf


def _make_fontfile_with_OS2(*, version, **kwargs):
	upem = 1000
	glyphOrder = [".notdef", "a"]
	cmap = {0x61: "a"}
	glyphs = {gn: Glyph() for gn in glyphOrder}
	hmtx = {gn: (500, 0) for gn in glyphOrder}
	names = {"familyName": "TestOS2", "styleName": "Regular"}

	fb = FontBuilder(unitsPerEm=upem)
	fb.setupGlyphOrder(glyphOrder)
	fb.setupCharacterMap(cmap)
	fb.setupGlyf(glyphs)
	fb.setupHorizontalMetrics(hmtx)
	fb.setupHorizontalHeader()
	fb.setupNameTable(names)
	fb.setupOS2(version=version, **kwargs)

	return _compile(fb.font)


def _merge_and_recompile(fontfiles, options=None):
	merger = Merger(options)
	merged = merger.merge(fontfiles)
	buf = _compile(merged)
	return ttLib.TTFont(buf)


@pytest.mark.parametrize(
	"v1, v2", list(itertools.permutations(range(5+1), 2))
)
def test_merge_OS2_mixed_versions(v1, v2):
	# https://github.com/fonttools/fonttools/issues/1865
	fontfiles = [
		_make_fontfile_with_OS2(version=v1),
		_make_fontfile_with_OS2(version=v2),
	]
	merged = _merge_and_recompile(fontfiles)
	assert merged["OS/2"].version == max(v1, v2)


if __name__ == "__main__":
	import sys
	sys.exit(unittest.main())



import logging
import os
import tempfile
import shutil
import unittest
from io import open
from .testSupport import getDemoFontGlyphSetPath
from fontTools.ufoLib.glifLib import (
	GlyphSet, glyphNameToFileName, readGlyphFromString, writeGlyphToString,
)
from fontTools.ufoLib.errors import GlifLibError, UnsupportedGLIFFormat, UnsupportedUFOFormat
from fontTools.misc.etree import XML_DECLARATION
from fontTools.pens.recordingPen import RecordingPointPen
import pytest

GLYPHSETDIR = getDemoFontGlyphSetPath()


class GlyphSetTests(unittest.TestCase):

	def setUp(self):
		self.dstDir = tempfile.mktemp()
		os.mkdir(self.dstDir)

	def tearDown(self):
		shutil.rmtree(self.dstDir)

	def testRoundTrip(self):
		import difflib
		srcDir = GLYPHSETDIR
		dstDir = self.dstDir
		src = GlyphSet(srcDir, ufoFormatVersion=2, validateRead=True, validateWrite=True)
		dst = GlyphSet(dstDir, ufoFormatVersion=2, validateRead=True, validateWrite=True)
		for glyphName in src.keys():
			g = src[glyphName]
			g.drawPoints(None)  # load attrs
			dst.writeGlyph(glyphName, g, g.drawPoints)
		# compare raw file data:
		for glyphName in sorted(src.keys()):
			fileName = src.contents[glyphName]
			with open(os.path.join(srcDir, fileName), "r") as f:
				org = f.read()
			with open(os.path.join(dstDir, fileName), "r") as f:
				new = f.read()
			added = []
			removed = []
			for line in difflib.unified_diff(
					org.split("\n"), new.split("\n")):
				if line.startswith("+ "):
					added.append(line[1:])
				elif line.startswith("- "):
					removed.append(line[1:])
			self.assertEqual(
				added, removed,
				"%s.glif file differs after round tripping" % glyphName)

	def testContentsExist(self):
		with self.assertRaises(GlifLibError):
			GlyphSet(
				self.dstDir,
				ufoFormatVersion=2,
				validateRead=True,
				validateWrite=True,
				expectContentsFile=True,
			)

	def testRebuildContents(self):
		gset = GlyphSet(GLYPHSETDIR, validateRead=True, validateWrite=True)
		contents = gset.contents
		gset.rebuildContents()
		self.assertEqual(contents, gset.contents)

	def testReverseContents(self):
		gset = GlyphSet(GLYPHSETDIR, validateRead=True, validateWrite=True)
		d = {}
		for k, v in gset.getReverseContents().items():
			d[v] = k
		org = {}
		for k, v in gset.contents.items():
			org[k] = v.lower()
		self.assertEqual(d, org)

	def testReverseContents2(self):
		src = GlyphSet(GLYPHSETDIR, validateRead=True, validateWrite=True)
		dst = GlyphSet(self.dstDir, validateRead=True, validateWrite=True)
		dstMap = dst.getReverseContents()
		self.assertEqual(dstMap, {})
		for glyphName in src.keys():
			g = src[glyphName]
			g.drawPoints(None)  # load attrs
			dst.writeGlyph(glyphName, g, g.drawPoints)
		self.assertNotEqual(dstMap, {})
		srcMap = dict(src.getReverseContents())  # copy
		self.assertEqual(dstMap, srcMap)
		del srcMap["a.glif"]
		dst.deleteGlyph("a")
		self.assertEqual(dstMap, srcMap)

	def testCustomFileNamingScheme(self):
		def myGlyphNameToFileName(glyphName, glyphSet):
			return "prefix" + glyphNameToFileName(glyphName, glyphSet)
		src = GlyphSet(GLYPHSETDIR, validateRead=True, validateWrite=True)
		dst = GlyphSet(self.dstDir, myGlyphNameToFileName, validateRead=True, validateWrite=True)
		for glyphName in src.keys():
			g = src[glyphName]
			g.drawPoints(None)  # load attrs
			dst.writeGlyph(glyphName, g, g.drawPoints)
		d = {}
		for k, v in src.contents.items():
			d[k] = "prefix" + v
		self.assertEqual(d, dst.contents)

	def testGetUnicodes(self):
		src = GlyphSet(GLYPHSETDIR, validateRead=True, validateWrite=True)
		unicodes = src.getUnicodes()
		for glyphName in src.keys():
			g = src[glyphName]
			g.drawPoints(None)  # load attrs
			if not hasattr(g, "unicodes"):
				self.assertEqual(unicodes[glyphName], [])
			else:
				self.assertEqual(g.unicodes, unicodes[glyphName])


class FileNameTest:

	def test_default_file_name_scheme(self):
		assert glyphNameToFileName("a", None) == "a.glif"
		assert glyphNameToFileName("A", None) == "A_.glif"
		assert glyphNameToFileName("Aring", None) == "A_ring.glif"
		assert glyphNameToFileName("F_A_B", None) == "F__A__B_.glif"
		assert glyphNameToFileName("A.alt", None) == "A_.alt.glif"
		assert glyphNameToFileName("A.Alt", None) == "A_.A_lt.glif"
		assert glyphNameToFileName(".notdef", None) == "_notdef.glif"
		assert glyphNameToFileName("T_H", None) =="T__H_.glif"
		assert glyphNameToFileName("T_h", None) =="T__h.glif"
		assert glyphNameToFileName("t_h", None) =="t_h.glif"
		assert glyphNameToFileName("F_F_I", None) == "F__F__I_.glif"
		assert glyphNameToFileName("f_f_i", None) == "f_f_i.glif"
		assert glyphNameToFileName("AE", None) == "A_E_.glif"
		assert glyphNameToFileName("Ae", None) == "A_e.glif"
		assert glyphNameToFileName("ae", None) == "ae.glif"
		assert glyphNameToFileName("aE", None) == "aE_.glif"
		assert glyphNameToFileName("a.alt", None) == "a.alt.glif"
		assert glyphNameToFileName("A.aLt", None) == "A_.aL_t.glif"
		assert glyphNameToFileName("A.alT", None) == "A_.alT_.glif"
		assert glyphNameToFileName("Aacute_V.swash", None) == "A_acute_V_.swash.glif"
		assert glyphNameToFileName(".notdef", None) == "_notdef.glif"
		assert glyphNameToFileName("con", None) == "_con.glif"
		assert glyphNameToFileName("CON", None) == "C_O_N_.glif"
		assert glyphNameToFileName("con.alt", None) == "_con.alt.glif"
		assert glyphNameToFileName("alt.con", None) == "alt._con.glif"

	def test_conflicting_case_insensitive_file_names(self, tmp_path):
		src = GlyphSet(GLYPHSETDIR)
		dst = GlyphSet(tmp_path)
		glyph = src["a"]

		dst.writeGlyph("a", glyph)
		dst.writeGlyph("A", glyph)
		dst.writeGlyph("a_", glyph)
		dst.deleteGlyph("a_")
		dst.writeGlyph("a_", glyph)
		dst.writeGlyph("A_", glyph)
		dst.writeGlyph("i_j", glyph)

		assert dst.contents == {
			'a': 'a.glif',
			'A': 'A_.glif',
			'a_': 'a_000000000000001.glif',
			'A_': 'A__.glif',
			'i_j': 'i_j.glif',
		}

		# make sure filenames are unique even on case-insensitive filesystems
		assert len({fileName.lower() for fileName in dst.contents.values()}) == 5


class _Glyph:
	pass


class ReadWriteFuncTest:

	def test_roundtrip(self):
		glyph = _Glyph()
		glyph.name = "a"
		glyph.unicodes = [0x0061]

		s1 = writeGlyphToString(glyph.name, glyph)

		glyph2 = _Glyph()
		readGlyphFromString(s1, glyph2)
		assert glyph.__dict__ == glyph2.__dict__

		s2 = writeGlyphToString(glyph2.name, glyph2)
		assert s1 == s2

	def test_xml_declaration(self):
		s = writeGlyphToString("a", _Glyph())
		assert s.startswith(XML_DECLARATION % "UTF-8")

	def test_parse_xml_remove_comments(self):
		s = b"""<?xml version='1.0' encoding='UTF-8'?>
		<!-- a comment -->
		<glyph name="A" format="2">
			<advance width="1290"/>
			<unicode hex="0041"/>
			<!-- another comment -->
		</glyph>
		"""

		g = _Glyph()
		readGlyphFromString(s, g)

		assert g.name == "A"
		assert g.width == 1290
		assert g.unicodes == [0x0041]

	def test_read_unsupported_format_version(self, caplog):
		s = """<?xml version='1.0' encoding='utf-8'?>
		<glyph name="A" format="0" formatMinor="0">
			<advance width="500"/>
			<unicode hex="0041"/>
		</glyph>
		"""

		with pytest.raises(UnsupportedGLIFFormat):
			readGlyphFromString(s, _Glyph())  # validate=True by default

		with pytest.raises(UnsupportedGLIFFormat):
			readGlyphFromString(s, _Glyph(), validate=True)

		caplog.clear()
		with caplog.at_level(logging.WARNING, logger="fontTools.ufoLib.glifLib"):
			readGlyphFromString(s, _Glyph(), validate=False)

		assert len(caplog.records) == 1
		assert "Unsupported GLIF format" in caplog.text
		assert "Assuming the latest supported version" in caplog.text

	def test_read_allow_format_versions(self):
		s = """<?xml version='1.0' encoding='utf-8'?>
		<glyph name="A" format="2">
			<advance width="500"/>
			<unicode hex="0041"/>
		</glyph>
		"""

		# these two calls are are equivalent
		readGlyphFromString(s, _Glyph(), formatVersions=[1, 2])
		readGlyphFromString(s, _Glyph(), formatVersions=[(1, 0), (2, 0)])

		# if at least one supported formatVersion, unsupported ones are ignored
		readGlyphFromString(s, _Glyph(), formatVersions=[(2, 0), (123, 456)])

		with pytest.raises(
			ValueError,
			match="None of the requested GLIF formatVersions are supported"
		):
			readGlyphFromString(s, _Glyph(), formatVersions=[0, 2001])

		with pytest.raises(GlifLibError, match="Forbidden GLIF format version"):
			readGlyphFromString(s, _Glyph(), formatVersions=[1])

	def test_read_ensure_x_y(self):
		"""Ensure that a proper GlifLibError is raised when point coordinates are
		missing, regardless of validation setting."""

		s = """<?xml version='1.0' encoding='utf-8'?>
		<glyph name="A" format="2">
			<outline>
				<contour>
					<point x="545" y="0" type="line"/>
					<point x="638" type="line"/>
				</contour>
			</outline>
		</glyph>
		"""
		pen = RecordingPointPen() 

		with pytest.raises(GlifLibError, match="Required y attribute"):
			readGlyphFromString(s, _Glyph(), pen)

		with pytest.raises(GlifLibError, match="Required y attribute"):
			readGlyphFromString(s, _Glyph(), pen, validate=False)

def test_GlyphSet_unsupported_ufoFormatVersion(tmp_path, caplog):
	with pytest.raises(UnsupportedUFOFormat):
		GlyphSet(tmp_path, ufoFormatVersion=0)
	with pytest.raises(UnsupportedUFOFormat):
		GlyphSet(tmp_path, ufoFormatVersion=(0, 1))


def test_GlyphSet_writeGlyph_formatVersion(tmp_path):
	src = GlyphSet(GLYPHSETDIR)
	dst = GlyphSet(tmp_path, ufoFormatVersion=(2, 0))
	glyph = src["A"]

	# no explicit formatVersion passed: use the more recent GLIF formatVersion
	# that is supported by given ufoFormatVersion (GLIF 1 for UFO 2)
	dst.writeGlyph("A", glyph)
	glif = dst.getGLIF("A")
	assert b'format="1"' in glif
	assert b'formatMinor' not in glif  # omitted when 0

	# explicit, unknown formatVersion
	with pytest.raises(UnsupportedGLIFFormat):
		dst.writeGlyph("A", glyph, formatVersion=(0, 0))

	# explicit, known formatVersion but unsupported by given ufoFormatVersion
	with pytest.raises(
		UnsupportedGLIFFormat,
		match="Unsupported GLIF format version .*for UFO format version",
	):
		dst.writeGlyph("A", glyph, formatVersion=(2, 0))
